#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#ifdef HAVE_SYS_CAPABILITY_H
# include <sys/capability.h>
#endif
#ifdef WITH_SECCOMP
# include <seccomp.h>
#endif

enum sc_mode {
	SC_FUSER = 1,
	SC_KILLALL,
	SC_PEEKFD,
	SC_PRTSTAT,
	SC_PSLOG,
	SC_PSTREE,
};

static void sandbox_install(enum sc_mode mode)
{
    /* Try to make descriptors 0-2 hopefully opened. */
    int in[2];
    if (pipe(in) == 0) {
	close(in[1]); /* Writing end. */
	int fd = in[0];
	while (fd != -1 && fd < 2)
	    fd = dup(fd);
	if (fd > 2)
	    close(fd);
    }

#ifndef __SANITIZE_ADDRESS__
    /* Ancient method of sandboxing. */
    struct rlimit rlim_zero = { 0, 0 };
    setrlimit(RLIMIT_NPROC, &rlim_zero);
    setrlimit(RLIMIT_FSIZE, &rlim_zero);
#endif

#ifdef HAVE_SYS_CAPABILITY_H
    /* Drop not-needed root capabilities. */
    cap_t caps = cap_init();
    if (caps) {
	cap_value_t cap_list[2];
	int ncap = 0;

	if (mode == SC_KILLALL || mode == SC_FUSER)
	    cap_list[ncap++] = CAP_KILL;
	/* Required to access /proc. */
	cap_list[ncap++] = CAP_SYS_PTRACE;

	cap_set_flag(caps, CAP_EFFECTIVE, ncap, cap_list, CAP_SET);
	cap_set_flag(caps, CAP_PERMITTED, ncap, cap_list, CAP_SET);
	cap_set_proc(caps);
	cap_free(caps);
    }
#endif

#ifdef WITH_SECCOMP
    scmp_filter_ctx ctx;

    /* No SCMP_ACT_KILL to make it (esp. killall) more robust. */
    ctx = seccomp_init(SCMP_ACT_ERRNO(EPERM));
    if (!ctx)
        return;

    /* Common syscalls. */
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(brk), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(clock_nanosleep), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(clock_nanosleep_time64), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(exit_group), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(exit), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(fcntl), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(fcntl64), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(fstat), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(fstat64), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(fstatat64), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(getcwd), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(getdents), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(getdents64), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(getpid), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(getsockname), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(lstat64), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(munmap), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(nanosleep), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(newfstatat), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(poll), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(ppoll), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(prlimit64), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(read), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(readlink), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(readlinkat), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(readv), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(rt_sigprocmask), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(rt_sigreturn), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(stat), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(stat64), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(statx), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(wait4), 0);
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(waitid), 0);

    /*
     * Allow read-only access for openat. openat2 cannot be
     * filtered for access type, thus, don't allow it at all.
     */
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(openat), 1,
		     SCMP_A2(SCMP_CMP_MASKED_EQ, O_ACCMODE, O_RDONLY));

    /* Do not allow write mmap/mprotect. This is redundant to open(2) filter. */
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(mmap), 1,
		     SCMP_A2(SCMP_CMP_MASKED_EQ, PROT_WRITE, 0));
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(mmap2), 1,
		     SCMP_A2(SCMP_CMP_MASKED_EQ, PROT_WRITE, 0));
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(mprotect), 1,
		     SCMP_A2(SCMP_CMP_MASKED_EQ, PROT_WRITE, 0));

    /* Allow TIOCGWINSZ for pstree. */
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(ioctl), 1,
		     SCMP_A0(SCMP_CMP_EQ, 1),
		     SCMP_A1(SCMP_CMP_EQ, TIOCGWINSZ));

    /* Signal senders. */
    if (mode == SC_KILLALL || mode == SC_FUSER)
	    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(kill), 0);

    if (mode == SC_PEEKFD)
	    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(ptrace), 0);

    /*
     * Filter similar to RLIMIT_FSIZE.
     * Do not allow closing stdio file descriptors, while allowing
     * writing to stdio file descriptors only.
     */
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(close), 1,
	SCMP_A0(SCMP_CMP_GT, 2));
    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(write), 1,
	SCMP_A0(SCMP_CMP_LE, 2));

    /* Allow some access to passwd. */
    if (mode == SC_KILLALL || mode == SC_FUSER || mode == SC_PSTREE) {
	    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(_llseek), 0);
	    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(lseek), 0);
	    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(lstat), 0);
	    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(connect), 0);
	    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(sendmsg), 0);
	    seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(socket), 1,
		SCMP_A0(SCMP_CMP_EQ, AF_UNIX));
    }

    seccomp_load(ctx);
    seccomp_release(ctx);
#endif
}

